"""Implementation of IIIF image manipulations using netpbm programs

FIXME - has not been kept up to date with other code
"""

import re
import os
import os.path
import glob
import subprocess

from error import IIIFError
from request import IIIFRequest
from manipulator import IIIFManipulator

class IIIFManipulatorNetpbm(IIIFManipulator):
    """Module to manipulate and image according to iiif rules

    All exceptions are raise as Error objects which directly
    determine the HTTP response.
    """

    tmpdir = '/tmp'
    filecmd = None
    pnmdir = None

    def __init__(self):
        super(IIIFManipulatorNetpbm, self).__init__()
        self.complianceLevel="http://iiif.example.org/compliance/level/1"
        if (self.filecmd is None):
            self.find_binaries()

    @classmethod
    def find_binaries(cls,tmpdir=None,shellsetup=None,filecmd=None,pnmdir=None):
        """Set instance vars for directory and binar locations

        FIXME - should accept params to set things other than defaults
        """
        cls.tmpdir = ( '/tmp' if (tmpdir is None) else tmpdir)
        # Shell setup command (e.g set library path)
        cls.shellsetup = ( '' if (shellsetup is None) else shellsetup)
        cls.filecmd = ( '/usr/bin/file' if (filecmd is None) else filecmd)
        if (pnmdir is None):
            cls.pnmdir = '/usr/bin'
            for dir in ('/usr/local/bin','/sw/bin'):
                if (os.path.isfile(os.path.join(dir,'pngtopnm'))):
                    cls.pnmdir=dir
        else:
            cls.pnmdir=pnmdir
        # Recklessly assume everything else under cls.pnmdir
        cls.pngtopnm  = os.path.join( cls.pnmdir , 'pngtopnm' )
        cls.pnmfile   = os.path.join( cls.pnmdir , 'pnmfile' )
        cls.pnmcut    = os.path.join( cls.pnmdir , 'pnmcut' )
        cls.pnmscale  = os.path.join( cls.pnmdir , 'pnmscale' )
        cls.pnmrotate = os.path.join( cls.pnmdir , 'pnmrotate' )
        cls.pnmflip   = os.path.join( cls.pnmdir , 'pnmflip' )
        cls.pnmtopng  = os.path.join( cls.pnmdir , 'pnmtopng' )
        cls.ppmtopgm  = os.path.join( cls.pnmdir , 'ppmtopgm' )
        cls.pnmtotiff = os.path.join( cls.pnmdir , 'pnmtotiff' )
        cls.pnmtojpeg = os.path.join( cls.pnmdir , 'pnmtojpeg' )
        cls.pamditherbw = os.path.join( cls.pnmdir , 'pamditherbw' )
        # Need djatoka to get jp2 output
        cls.djatoka_comp = '/Users/simeon/packages/adore-djatoka-1.1/bin/compress.sh'

    def do_first(self):
        pid = os.getpid()
        self.basename = os.path.join(self.tmpdir,'iiif_netpbm_'+str(pid))
        outfile = self.basename+'.pnm'
        # Convert source file to pnm
        filetype = self.file_type(self.srcfile)
        if (filetype == 'png'):
            #print "doing png->pnm"
            #print self.pngtopnm+' '+self.srcfile+' > '+outfile
            if (self.shell_call(self.pngtopnm+' '+self.srcfile+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pngtopnm.")
        else:
            raise IIIFError(code='501',
                           text='bad input file format')
        self.tmpfile = outfile

    def do_region(self):
        infile = self.tmpfile
        outfile = self.basename+'.reg'
        # Get size
        (self.width,self.height)=self.image_size(infile)
        # Region
        #simeon@ice ~>cat m.pnm | pnmcut 10 10 100 200 > m1.pnm
        (x,y,w,h)=self.region_to_apply()
        if (x is None):
            #print "region: full"
            self.tmpfile = infile
        else:
            #print "region: (%d,%d,%d,%d)" % (x,y,w,h)
            if (self.shell_call('cat '+infile+' | '+self.pnmcut+' '+str(x)+' '+str(y)+' '+str(w)+' '+str(h)+'  > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmcut.")
            self.width=w
            self.height=h
            self.tmpfile = outfile

    def do_size(self):
        # Size
        # simeon@ice ~>cat m1.pnm | pnmscale -width 50 > m2.pnm
        infile = self.tmpfile
        outfile = self.basename+'.siz'
        (w,h)=self.size_to_apply()
        if (w is None):
            #print "size: no scaling"
            self.tmpfile=infile
        else:
            #print "size: scaling to (%d,%d)" % (w,h)
            if (self.shell_call('cat '+infile+' | '+self.pnmscale+' -width '+str(w)+' -height '+str(h)+'  > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmscale.")
            width=w
            height=h
            self.tmpfile=outfile

    def do_rotation(self):
        infile = self.tmpfile
        outfile = self.basename+'.rot'
        # Rotate
        #simeon@ice ~>cat m2.pnm | pnmrotate 123.456 > m3.pnm
        #pnmrotate: angle must be between -90 and 90 and it is CCW not CW
        rot=self.rotation_to_apply()
        if (rot==0.0):
            #print "rotation: no rotation"
            self.tmpfile=infile
        elif (rot<=90.0 or rot>=270.0):
            if (rot>=270.0):
                rot-=360.0
            #print "rotation: by %f degrees clockwise" % (rot)
            if (self.shell_call('cat '+infile+' | '+self.pnmrotate+' -background=#FFF '+str(-rot)+'  > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmrotate.")
            self.tmpfile=outfile
        else:
            # Between 90 and 270 = flip and then -90 to 90
            rot-=180.0
            #print "rotation: by %f degrees clockwise" % (rot)
            if (self.shell_call('cat '+infile+' | '+self.pnmflip+' -rotate180 | '+self.pnmrotate+' '+str(-rot)+'  > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmrotate.")
            self.tmpfile=outfile

    def do_quality(self):
        infile = self.tmpfile
        outfile = self.basename+'.col'
        # Quality (bit-depth):
        quality=self.quality_to_apply()
        if (quality == 'grey'):
            #print "quality: grey"
            if (self.shell_call('cat '+infile+' | '+self.ppmtopgm+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from ppmtopgm.")
            self.tmpfile=outfile
        elif (quality == 'bitonal'):
            #print "quality: grey"
            if (self.shell_call('cat '+infile+' | '+self.ppmtopgm+' | '+self.pamditherbw+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from ppmtopgm.")
            self.tmpfile=outfile
        elif (quality == 'quality'):
            #print "quality: quality"
            self.tmpfile=infile
        else: 
            raise IIIFError(code=400,parameter='quality',
                           text="Unknown quality parameter value requested.")

    def do_format(self):
        infile = self.tmpfile
        outfile = self.basename+'.out'
        outfile_jp2 = self.basename+'.jp2'
        # Now convert finished pnm file to output format
        #simeon@ice ~>cat m3.pnm | pnmtojpeg  > m4.jpg
        #simeon@ice ~>cat m3.pnm | pnmtotiff > m4.jpg
        #pnmtotiff: computing colormap...
        #pnmtotiff: Too many colors - proceeding to write a 24-bit RGB file.
        #pnmtotiff: If you want an 8-bit palette file, try doing a 'ppmquant 256'.
        #simeon@ice ~>cat m3.pnm | pnmtopng  > m4.png
        fmt = ( 'png' if (self.request.format is None) else self.request.format)
        if (fmt == 'png'):
            #print "format: png"
            if (self.shell_call(self.pnmtopng+' '+infile+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmtopng.")
            mime_type="image/png"
        elif (fmt == 'tiff' or fmt == 'jp2'):
            #print "format: tiff/jp2"
            if (self.shell_call(self.pnmtotiff+' '+infile+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmtotiff.")
            mime_type="image/tiff"
            if (fmt == 'jp2'):
                # use djatoka after tiff
                if (self.shell_call(DJATOKA_COMP+' -i '+outfile+' -o '+outfile_jp2)):
                    raise IIIFError(text="Oops... got nonzero output from DJATOKA_COMP.")
                mime_type="image/jp2"
                outfile=tmpfile_jp2
        elif (fmt == 'jpg'):
            #print "format: jpg"
            if (self.shell_call(self.pnmtojpeg+' '+infile+' > '+outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmtojpeg.")
            mime_type="image/jpeg"
        else:
            raise IIIFError(code=415, parameter='format',
                           text="Unsupported output file format (%s), only png,jpg,tiff are supported."%(fmt))
        self.outfile=outfile
        self.output_format=fmt
        self.mime_type=mime_type

    def file_type(self,file):
        return('png')
        pout = os.popen(self.filecmd+' '+file,'r')
        fileout=pout.read(200)
        pout.close()
        #print "file output = " + fileout
        if (re.search('PNG image data',fileout)):
            return('png')
        # failed
        return(None)

    def image_size(self,pnmfile):
        """Get width and height of pnm file

        simeon@homebox src>pnmfile /tmp/214-2.png
        /tmp/214-2.png:PPM raw, 100 by 100  maxval 255
        """
        pout = os.popen(self.shellsetup+self.pnmfile+' '+pnmfile,'r')
        pnmfileout=pout.read(200)
        pout.close()
        m=re.search(', (\d+) by (\d+) ',pnmfileout);
        if (m is None):
            raise IIIFError(text="Bad output from pnmfile when trying to get size.")
        w=int(m.group(1))
        h=int(m.group(2))
        #print "pnmfile output = %s" % (pnmfileout)
        #print "image size = %d,%d" % (w,h)
        return(w,h)

    def shell_call(self,shellcmd):
        """Shell call with necessary setup first"""
        return(subprocess.call(self.shellsetup+shellcmd,shell=True))

    def cleanup(self):
        """Clean up any temporary files"""
        for file in glob.glob(self.basename+'*'):
            os.unlink(file)



"""IIIF error class

<?xml version="1.0" encoding="UTF8"?>
<error xmlns="http://library.stanford.edu/iiif/image-api/ns/">
  <parameter>size</parameter>
  <text>Invalid size specified</text>
</error>
"""

import sys
import StringIO
from xml.etree.ElementTree import ElementTree, Element

# Namespace used in XML error response
I3F_NS = "http://library.stanford.edu/iiif/image-api/ns/"

class IIIFError:

    def __init__(self,code=500,parameter='unknown',text=None):
        self.code=code
        self.parameter=parameter
        self.text=text
        self.pretty_xml=True

    def __repr__(self):
        """XML representation of the error to be used in HTTP response
        """
        # Build tree
        spacing = ( "\n" if (self.pretty_xml) else "" )
        root = Element('error', { 'xmlns': I3F_NS } )
        root.text=spacing
        e_parameter = Element( 'parameter', {} )
        e_parameter.text = self.parameter
        e_parameter.tail = spacing
        root.append(e_parameter)
        if (self.text is not None):
            e_text = Element( 'text', {} )
            e_text.text = self.text
            e_text.tail = spacing
            root.append(e_text)
        
        # Write out as XML document to return
        tree = ElementTree(root);
        xml_buf=StringIO.StringIO()
        if (sys.version_info < (2,7)):
            tree.write(xml_buf,encoding='UTF-8')
        else:
            tree.write(xml_buf,encoding='UTF-8',xml_declaration=True,method='xml')
        return(xml_buf.getvalue())



"""IIIF Image Information Response

Model for info.json. The Image Information Response has been
in just the JSON format, with JSOD-LD extensions, since v1.1
of the Image API.
"""

import sys
import json
import re
import StringIO

# JSON-LD context
# 1.1
PARAMS = ['identifier','protocol','width','height','scale_factors','tile_width','tile_height','formats','qualities','profile']
EVAL_PARAMS = set(['scale_factors','formats','qualities'])

class IIIFInfo(object):

    def __init__(self,level=1, identifier=None,width=None,height=None,
                 scale_factors=None,tile_width=None,tile_height=None,
                 formats=None,qualities=None,api_version='2.0',
                 conf=None,):
        # API version (used in level settings)
        self.api_version = api_version
        self.set_version_info()
        # explicit settings
        self.identifier = identifier
        self.width = width
        self.height = height
        self.scale_factors = scale_factors
        self.tile_width = tile_width
        self.tile_height = tile_height
        self.formats = formats
        self.qualities = qualities
        self.level = level
        # defaults from conf dict if provided
        if (conf):
            for option in conf:
                if (option in EVAL_PARAMS):
                    self.__dict__[option]=eval(conf[option]) #FIXME - avoid eval
                else:
                    self.__dict__[option]=conf[option]

    def set_version_info(self):
        if (self.api_version=='1.1'):
            self.context = "http://library.stanford.edu/iiif/image-api/1.1/context.json"
            self.profile_prefix = "http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level"
            self.profile_suffix = ""
            self.protocol = None
            self.required_params = ['identifier','width','height','profile']
        elif (self.api_version=='2.0'):
            self.context = "http://iiif.io/api/image/2.0/context.json"
            self.profile_prefix = "http://iiif.io/api/image/2/level"
            self.profile_suffix = ".json"
            self.protocol = "http://iiif.io/api/image"
            self.required_params = ['identifier','width','height','protocol','profile']

    @property
    def level(self):
        """Extract level number from profile URI

        Returns integer level number or raises excpetion
        """
        m = re.match(self.profile_prefix+r'(\d)'+self.profile_suffix+"$",self.profile)
        if (m):
            return int(m.group(1))
        raise Exception("Bad compliance profile URI, failed to extract level number")

    @level.setter
    def level(self, value):
        """Build profile URI from level
        
        Level should be an integer 0,1,2
        """
        self.profile = self.profile_prefix + ("%d" % value) + self.profile_suffix

    def set(self,param,value):
        if (param in EVAL_PARAMS):
            self.__dict__[param]=eval(value) #FIXME - avoid eval
        else:
            self.__dict__[param]=value

    def validate(self):
        """Validate this object as Image API data

        Raise Exception with helpful message if not valid.
        """
        errors = []
        for param in self.required_params:
            if (param not in self.__dict__ or
                self.__dict__[param] is None):
                errors.append("missing %s parameter" % (param))
        if (len(errors)>0):
            raise Exception("Bad data for inso.json: "+", ".join(errors))
        return True

    def as_json(self, validate=True):
        """Return JSON serialization
        
        Will raise exception if insufficient parameters are present to
        have a valid info.json response (unless validate is False).
        """
        if (validate):
            self.validate()
        json_dict = {}
        json_dict['@context']=self.context
        if (self.identifier):
            json_dict['@id']=self.identifier
        for param in PARAMS:
            if (param in self.__dict__ and 
                self.__dict__[param] is not None and
                param!='identifier'):
                json_dict[param] = self.__dict__[param]
        return( json.dumps(json_dict, sort_keys=True, indent=2 ) )



from _version import __version__

from config import IIIFConfig
from error import IIIFError
from info import IIIFInfo
from manipulator import IIIFManipulator
from request import IIIFRequest
from static import IIIFStatic
#iiif/manipulator_netpbm.py
#iiif/manipulator_pil.py



"""
Implementation of IIIF Image API manipulations using the Python Image Library

Uses the Python Image Library (PIL) for in-python manipulation:
http://www.pythonware.com/products/pil/index.htm
"""

import re
import os
import os.path
import subprocess
import tempfile

from PIL import Image

from error import IIIFError
from request import IIIFRequest
from manipulator import IIIFManipulator

class IIIFManipulatorPIL(IIIFManipulator):
    """Module to manipulate and image according to iiif rules

    All exceptions are raise as Error objects which directly
    determine the HTTP response.
    """

    tmpdir = '/tmp'
    filecmd = None
    pnmdir = None

    def __init__(self):
        super(IIIFManipulatorPIL, self).__init__()
        # Does not support jp2 output
        self.complianceLevel="http://iiif.example.org/compliance/level/0"
        self.outtmp = None

    def do_first(self):
        """Create PIL object from input image file
        """
        self.logger.info("do_first: src=%s" % (self.srcfile))
        try:
            self.image=Image.open(self.srcfile)
            self.image.load()
        except Exception as e:
            raise IIIFError(text=("PIL Image.open(%s) barfed: %s",(self.srcfile,str(e))))
        (self.width,self.height)=self.image.size

    def do_region(self):
        (x,y,w,h)=self.region_to_apply()
        if (x is None):
            self.logger.info("region: full (nop)")
        else:
            self.logger.info("region: (%d,%d,%d,%d)" % (x,y,w,h))
            self.image = self.image.crop( (x,y,x+w,y+h) )
            self.width = w
            self.height = h

    def do_size(self):
        (w,h)=self.size_to_apply()
        if (w is None):
            self.logger.info("size: no scaling (nop)")
        else:
            self.logger.info("size: scaling to (%d,%d)" % (w,h))
            self.image = self.image.resize( (w,h) )
            self.width = w
            self.height = h

    def do_rotation(self):
        rot=self.rotation_to_apply()
        if (rot==0.0):
            self.logger.info("rotation: no rotation (nop)")
        else:
            self.logger.info("rotation: by %f degrees clockwise" % (rot))
            self.image = self.image.rotate( -rot, expand=True )

    def do_quality(self):
        quality=self.quality_to_apply()
        if (quality == 'grey'):
            self.logger.info("quality: grey")
        elif (quality == 'bitonal'):
            self.logger.info("quality: bitonal")
        else:
            self.logger.info("quality: quality (nop)")

    def do_format(self):
        # assume tiling apps want jpg...
        fmt = ( 'jpg' if (self.request.format is None) else self.request.format)
        if (fmt == 'png'):
            self.logger.info("format: png")
            self.mime_type="image/png"
            self.output_format=fmt
            format = 'png'
        elif (fmt == 'jpg'):
            self.logger.info("format: jpg")
            self.mime_type="image/jpeg"
            self.output_format=fmt
            format = 'jpeg';
        else:
            raise IIIFError(code=415, parameter='format',
                           text="Unsupported output file format (%s), only png,jpg are supported."%(fmt))

        if (self.outfile is None):
            # Create temp
            f = tempfile.NamedTemporaryFile(delete=False)
            self.outfile=f.name
            self.outtmp=f.name
            self.image.save(f,format='png')
        else:
            # Save to specified location
            self.image.save(self.outfile,format='jpeg')

    def cleanup(self):
        if (self.outtmp is not None):
            try:
                os.unlink(self.outtmp)
            except OSError as e:
                # FIXME - should log warning when we have logging...
                pass



""" Create and parse IIIF request URLs

This class is a thorough implementation of restrictions in the
IIIF specification. It does not add any implementation specific 
restrictions.
"""
import urllib
import re
import string

from error import IIIFError

class IIIFRequest:
    """ Implement IIIF request URL syntax
    
    There are two URL forms defined in section 2:
    http[s]://server/[prefix/]identifier/region/size/rotation/quality[.format]
    or
    http[s]://server/[prefix/]identifier/info.format

    The attribites of objects of this class follow the names here except that
    baseurl is used for "http[s]://server/[prefix/]". If baseurl is not set 
    then a relative URL will be created or parsed.
    """

    def __init__(self, **params):
        """Create Request object and optionally set any attributes via 
        named parameters

        Unless specified the baseurl will be set to nothing ("").
        """
        self.clear()
        self.region_full=False
        self.region_pct=False
        self.region_xywh=None # (x,y,w,h)
        self.size_wh=None     # (w,h)
        #
        self.baseurl = ''
        self.set(**params)

    def clear(self):
        """ Clear all data that might pertain to an individual iiif URL

        Does not change/reset the baseurl which might be useful in a
        sequence of calls.
        """
        self.identifier = None
        self.region = None
        self.size = None
        self.rotation = None
        self.quality = None
        self.format = None
        self.info = None
        # Other flags
        self.size_full = False
        self.size_pct = None
        self.size_bang = None
        self.rotation_deg = 0.0 

    def set(self, **params):
        # FIXME - maybe make this safe an allow only setting valid attributes
        self.__dict__.update(**params)

    def quote(self, path_segment):
        """ Quote parameters in IIIF URLs

        Quote by percent-encoding "%" and gen-delims of RFC3986 exept colon
        
        http://tools.ietf.org/html/rfc3986#section-2.3
        gen-delims  = ":" / "/" / "?" / "#" / "[" / "]" / "@"
         
        to-encode = "%" / "/" / "?" / "#" / "[" / "]" / "@"
        """
        return( urllib.quote(path_segment,"-._~!$&'()*+,;=:") ) #FIXME - quotes too much

    def url(self, **params):
        """ Build a URL path according to the IIIF API parameterized or 
            info form

        The parameterized form is assumed unless a the info parameter is 
        specified.
        """
        self.set(**params)
        path = self.baseurl +\
            self.quote(self.identifier) + "/"
        if (self.info):
            # info request
            path += "info"
            format = self.format if self.format else "json"
        else:
            # set defaults if not given
            if self.region:
                region = self.region
            elif self.region_xywh:
                region = "%d,%d,%d,%d" % (self.region_xywh[0],self.region_xywh[1],self.region_xywh[2],self.region_xywh[3]) #FIXME - how to get tiple from list
            else:
                region = "full"
            if self.size:
                size = self.size
            elif self.size_wh:
                size = "%d,%d" % (self.size_wh[0],self.size_wh[1])
            else:
                size = "full" 
            rotation = self.rotation if self.rotation else "0"
            quality = self.quality if self.quality else "native"
            # parameterized form
            path += self.quote(region) + "/" +\
                    self.quote(size) + "/" +\
                    self.quote(rotation) + "/" +\
                    self.quote(quality)
            format = self.format
        if (format):
            path += "." + format
        return(path)

    def parse_url(self, url):
        """ Parse an IIIF API URL path and each component
        
        Will parse a URL or URL path that accords with either the
        parametrized or info request forms. Will raise an
        IIIFError on failure. A wrapper for the split_url()
        and parse_parameters() methods.
        """
        self.split_url(url)
        if (not self.info):
            self.parse_parameters()
        return(self)

    def split_url(self,url):
        """ Perform the initial parsing of an IIIF API URL path into components

        Will parse a URL or URL path that accords with either the
        parametrized or info API forms. Will raise an IIIFError on 
        failure.
        """
        # clear data first
        self.clear()
        # url must start with baseurl if set
        if (self.baseurl): 
            (path, num) = re.subn('^'+self.baseurl, '', url, 1)
            if (num != 1):
                raise(IIIFError("URL does not match baseurl (server/prefix)."))
            url=path
        # Look for optional .fmt at end, must start with letter. Note that we 
        # want to catch the case of a dot and no format (format='') which is
        # different from no dot (format=None)
        m = re.match( "(.+)\.([a-zA-Z]\w*)$", url)
        if (m):
            # There is a format string at end, chop off and store
            url = m.group(1)
            self.format = ( m.group(2) if (m.group(2) is not None) else '')
        # Break up by path segments, count to decide format
        segs = string.split( url, '/', 5)
        if (len(segs) > 5):
            raise(IIIFError(code=400,text="Too many path segments in URL (got %d: %s) in URL."%(len(segs),' | '.join(segs))))
        elif (len(segs) == 5):
            self.identifier = urllib.unquote(segs[0])
            self.region = urllib.unquote(segs[1])
            self.size  = urllib.unquote(segs[2])
            self.rotation = urllib.unquote(segs[3])
            self.quality = urllib.unquote(segs[4])
            self.info = False
        elif (len(segs) == 2):
            self.identifier = urllib.unquote(segs[0])
            if (urllib.unquote(segs[1]) != "info"):
                raise(IIIFError(code=400,text="Badly formed information request, must be info.json or info.xml"))
            if (self.format not in ("json","xml")):
                raise(IIIFError(code=400,text="Bad information request format, must be json or xml"))
            self.info = True
        else:
            raise(IIIFError(code=400,text="Bad number of path segments (%d: %s) in URL."%(len(segs),' | '.join(segs))))
        return(self)

    def parse_parameters(self):
        """ Parse the parameters of a parameterized request

        Will throw an IIIFError on failure, set attributes on success. Care is 
        take not to change any of the artibutes which store path components, 
        all parsed values are stored in new attributes.
        """
        self.parse_region()
        self.parse_size()
        self.parse_rotation()
        self.parse_quality()
        self.parse_format()

    def parse_region(self):
        """ Parse the region component of the path

        /full/ -> self.region_full = True (test this first)
        /x,y,w,h/ -> self.region_xywh = (x,y,w,h)
        /pct:x,y,w,h/ -> self.region_xywh and self.region_pct = True

        Will throw errors if the paremeters are illegal according to the
        specification but does not know about and thus cannot do any tests
        against any image being manipulated.
        """
        self.region_full=False
        self.region_pct=False
        if (self.region is None or self.region == 'full'):
            self.region_full=True
            return
        xywh=self.region
        pct_match = re.match('pct:(.*)$',self.region)
        if (pct_match):
            xywh=pct_match.group(1)
            self.region_pct=True
        # Now whether this was pct: or now, we expect 4 values...
        str_values = string.split(xywh, ',', 5)
        if (len(str_values) != 4):
            raise IIIFError(code=400,parameter="region",
                           text="Bad number of values in region specification, must be x,y,w,h but got %d value(s) from '%s'"%(len(str_values),xywh))
        values=[]
        for str_value in str_values:
            # Must be either integer (not pct) or interger/float (pct)
            if (pct_match):
                try:
                    # This is rather more permissive that the iiif spec
                    value=float(str_value)
                except ValueError:
                    raise IIIFError(code=400,parameter="region",
                                   text="Bad floating point value for percentage in region (%s)."%str_value)
                if (value>100.0):
                    raise IIIFError(code=400,parameter="region",
                                   text="Percentage over value over 100.0 in region (%s)."%str_value)
            else:
                try:
                    value=int(str_value)
                except ValueError:
                    raise IIIFError(code=400,parameter="region",
                                   text="Bad integer value in region (%s)."%str_value)
            if (value<0):
                raise IIIFError(code=400,parameter="region",
                               text="Negative values not allowed in region (%s)."%str_value)
            values.append(value)
        # Zero size region is w or h are zero (careful that they may be float)
        if (values[2]==0.0 or values[3]==0.0):
            raise IIIFError(code=400,parameter="region",
                           text="Zero size region specified (%s))."%xywh)
        self.region_xywh=values

    def parse_size(self):
        """Parse the size component of the path

        /full/ -> self.size_full = True
        /w,/ -> self.size_wh = (w,None)
        /,h/ -> self.size_wh = (None,h)
        /w,h/ -> self.size_wh = (w,h)
        /pct:p/ -> self.size_pct = p
        /!w,h/ -> self.size_wh = (w,h), self.size_bang = True

        Expected use:
          (w,h) = iiif.size_to_apply(region_w,region_h)
          if (q is None):
              # full image
          else:
              # scale to w by h
        Returns (None,None) if no scaling is required.
        """
        self.size_pct=None
        self.size_bang=False
        self.size_full=False
        self.size_wh=(None,None)
        if (self.size is None or self.size=='full'):
            self.size_full=True
            return
        pct_match = re.match('pct:(.*)$',self.size)
        if (pct_match is not None):
            pct_str=pct_match.group(1)
            try:
                self.size_pct=float(pct_str)
            except ValueError:
                raise IIIFError(code=400,parameter="size",
                               text="Percentage size value must be a number, got '%s'."%(pct_str))
# FIXME - current spec places no upper limit on size
#            if (self.size_pct<0.0 or self.size_pct>100.0):
#                raise IIIFError(code=400,parameter="size",
#                               text="Illegal percentage size, must be 0 <= pct <= 100.")
            if (self.size_pct<0.0):
                raise IIIFError(code=400,parameter="size",
                               text="Base size percentage, must be > 0.0, got %f."%(self.size_pct))
        else:
            bang_match = re.match('!(.*)$',self.size)
            if (bang_match is not None):
                # Have "!w,h" form
                (mw,mh)=self._parse_w_comma_h(bang_match.group(1),'size')
                if (mw is None or mh is None):
                    raise IIIFError(code=400,parameter="size",
                                   text="Illegal size requested: both w,h must be specified in !w,h requests.")
                self.size_bang=True
            else:
                # Must now be "w,h", "w," or ",h"
                self.size_wh=self._parse_w_comma_h(self.size,'size')
            # Sanity check w,h
            (w,h) = self.size_wh
            if ( ( w is not None and w<=0) or 
                 ( h is not None and h<=0) ):
                raise IIIFError(code=400,parameter='size',
                               text="Size parameters request zero size result image.")

    def _parse_w_comma_h(self,str,param):
        """ Utility to parse "w,h" "w," or ",h" values
        
        Returns (w,h) where w,h are either None or ineteger. Will
        throw a ValueError if there is a problem with one or both.
        """
        (wstr,hstr) = string.split(str, ',', 2)
        try:
            w = self._parse_non_negative_int(wstr,'w')
            h = self._parse_non_negative_int(hstr,'h')
        except ValueError as e:
            raise IIIFError(code=400,parameter=param,
                           text="Illegal parameter value (%s)." % str(e) )
        if (w is None and h is None):
            raise IIIFError(code=400,parameter=param,
                           text="Must specify at least one of w,h.")
        self.size_wh=(w,h)
        return(w,h)

    def _parse_non_negative_int(self,istr,name):
        """ Parse integer from string (istr)

        The (name) parameter is used just for IIIFError message generation
        to indicate what the error is in.
        """
        if (istr == ''):
            return(None)
        try:
            i = int(istr)
        except ValueError:
            raise ValueError("Failed to extract integer value for %s" % (name))
        if (i<0):
            raise ValueError("Illegal negative value for %s" % (name))
        return(i)

    def parse_rotation(self):
        """ Check and interpret rotation

        Sets self.rotation_deg to a floating point number 0 <= angle < 360. Includes
        translation of 360 to 0.
        """
        if (self.rotation is None):
            self.rotation_deg=0.0
            return
        try:
            self.rotation_deg=float(self.rotation)
        except ValueError:
            raise IIIFError(code=400,parameter="rotation",
                           text="Bad rotation value, must be a number, got '%s'."%(self.rotation))
        if (self.rotation_deg<0.0 or self.rotation_deg>360.0):
            raise IIIFError(code=400,parameter="rotation",
                           text="Illegal rotation value, must be 0 <= rotation <= 360, got %f."%(self.rotation_deg))
        elif (self.rotation_deg==360.0):
            # The spec admits 360 as valid, but change to 0
            self.rotation_deg=0.0

    def parse_quality(self):
        """ Check quality paramater

        Sets self.quality_val based on simple substitution of 'native' for 
        default. Checks for the three valid values else throws and IIIFError.
        """
        if (self.quality is None):
            self.quality_val='native'
        elif (self.quality not in ['native','color','bitonal','grey']):
            raise IIIFError(code=400,parameter="quality",
                           text="The quality parameter must be 'native', 'color', 'bitonal' or 'grey', got '%s'."%(self.quality))
        else:
            self.quality_val=self.quality

    def parse_format(self):
        """ Check format parameter

        FIXME - do something...
        """
        pass

    def __str__(self):
        """ Pretty print this object in human readable form
        
        Distinguishes parametrerized and info requests to
        show only appropriate parameters in each case.
        """
        s =  "baseurl=" + str(self.baseurl) + " "
        s += "identifier=" + str(self.identifier) + " "
        if (self.info):
            s += "INFO request ";
            s += "format=" + str(self.format) + " "
        else:
            s += "region=" + str(self.region) + " "
            s += "size=" + str(self.size) + " "
            s += "rotation=" + str(self.rotation) + " "
            s += "quality=" + str(self.quality) + " "
            s += "format=" + str(self.format)
        return(s)



"""iiif configs

Goal is to have a sane set of default configs so that even if there
isn't entries in iiif.conf, or there are parameters missing, then something
useful will still happen.
"""
import ConfigParser
import os

I3F_CONF = 'etc/iiif.conf'

DEFAULTS = {
    'info' : {
        'tile_height': '256',
        'tile_width': '256',
        'scale_factors': '[1,2,4,8]',
        'formats': '[ "jpg", "png" ]',
        'qualities': '[ "native", "color" ]',
        },
    'test' : {
        'server_host': '',
        'server_port': '8000',
        'image_dir': 'testimages',
        'run_dummy': '1',
        'dummy_prefix': 'dummy',
        'run_pil': '1',
        'pil_prefix': 'pil',
        'run_netpbm': '',
        'netpbm_prefix': 'netpbm',
        },
}

class IIIFConfig(object):

    def __init__(self,conf_file=None):
        if (conf_file is None):
            root = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
            conf_file = os.path.join(root, I3F_CONF)
        self.conf = ConfigParser.ConfigParser()
        # set defaults
        for section in DEFAULTS.keys():
            self.conf.add_section(section)
            for option in DEFAULTS[section].keys():
                self.conf.set(section,option,DEFAULTS[section][option])
        # read config if present
        self.conf.read(conf_file)

    def get(self, section, option, raw=False, vars=None):
        """Wrapper for get(..) in ConfigParser"""
        return self.conf.get(section,option,raw,vars)



""" Almost null implementation iiif image manipulations and base class 

Provides a number of utility methods to extract information necessary
for doing the transformations once one has knowledge of the source
image size.
"""

import logging
import os
import os.path
import re
import shutil
import subprocess

from error import IIIFError
from request import IIIFRequest

class IIIFManipulator(object):
    """ Manipulate an image according to IIIF rules

    All exceptions are raise as IIIFError objects which directly
    determine the HTTP response.
    """

    def __init__(self):
        """ Return URI indicating IIIF compliance level, or None if not

        The attribute complicanceLevel is set of either None or a URI 
        that can be used in the HTTP response headers like of the form

        Link: <http://iiif.example.org/compliance/level/0>;rel="compliesTo"

        This null manipulator doesn't comply to any of the levels defined
        in the iiif specification so it is set to None.
        """
        self.srcfile = None
        self.request = None
        self.outfile = None
        self.complianceLevel = None
        self.logger = logging.getLogger(__name__)

    def derive(self,srcfile=None,request=None,outfile=None):
        """ Do sequence of manipulations for IIIF to derive output image
        
        Args:
            srcfile - source image file
            request - IIIFRequest object with parsed parameters
            outfile - output image file. If set the the output file will be
                      written to that file, otherwise a new temporary file
                      will be created and outfile set to its location.

        See order in spec: http://www-sul.stanford.edu/iiif/image-api/#order

          Region THEN Size THEN Rotation THEN Quality THEN Format

        Typical use:
            
            r = IIIFRequest(region=...)
            m = IIIFManipulator()
            try:
                m.derive(srcfile='a.jpg',request=r)
                # .. serve m.outfile
            except IIIFError as e:
                # ..
            finally:
                m.cleanup() #removes temp m.outfile

        """
        # set if specified
        if (srcfile is not None):
            self.srcfile=srcfile
        if (request is not None):
            self.request=request
        if (outfile is not None):
            self.outfile=outfile
        if (self.outfile is not None):
            # create path to output dir if necessary
            dir = os.path.dirname(self.outfile)
            if (not os.path.exists(dir)):
                os.makedirs(dir)
        #
        self.do_first()
        self.do_region()
        self.do_size()
        self.do_rotation()
        self.do_quality()
        self.do_format()
        self.do_last()
        return(self.outfile,self.mime_type)

    def do_first(self):
        """Simplest possible manipulator that can only handle no modification

        Set width and height to -1 (unknown)
        """
        self.width=-1  #don't know width of height
        self.height=-1 

    def do_region(self):
        # Region
        (x,y,w,h)=self.region_to_apply()
        if (x is not None):
            raise IIIFError(code=501,parameter="region",
                            text="Null manipulator supports only region=/full/.")

    def do_size(self):
        # Size
        # (w,h) = self.size_to_apply()
        if (self.request.size_pct != 100.0 and
            self.request.size != 'full'):
            raise IIIFError(code=501,parameter="size",
                            text="Null manipulator supports only size=pct:100 and size=full.")

    def do_rotation(self):
        # Rotate
        if (self.rotation_to_apply() != 0.0):
            raise IIIFError(code=501,parameter="rotation",
                            text="Null manipulator supports only rotation=(0|360).")

    def do_quality(self):
        # Quality
        if (self.quality_to_apply() != "native"):
            raise IIIFError(code=501,parameter="native",
                            text="Null manipulator supports only color=color.")

    def do_format(self):
        # Format (the last step)
        if (self.request.format is not None):
            raise IIIFError(code=415,parameter="format",
                            text="Null manipulator does not support specification of output format.")
        # 
        if (self.outfile is None):
            self.outfile=self.srcfile
        else:
            try:
                shutil.copyfile(self.srcfile,self.outfile)
            except IOError as e:
                raise IIIFError(code=500,
                                text="Failed to copy file (%s)." % (str(e)))
        self.mime_type=None


    def do_last(self):
        """ Hook in pipeline at end of processing

        Does nothing.
        """
        return

    ### Utility methods

    def region_to_apply(self):
        """Return the x,y,w,h parameters to extract given image width and height

        Assume image width and height are available in self.width and 
        self.height, and self.request is IIIFRequest object 

        Expected use:
          (x,y,w,h) = self.region_to_apply()
          if (x is None):
              # full image
          else:
              # extract
        Returns (None,None,None,None) if no extraction is required.
        """
        if (self.request.region_full):
            return(None,None,None,None)
        # Cannot do anything else unless we know size (in self.width and self.height)
        if (self.width<=0 or self.height<=0):
            raise IIIFError(code=501,parameter='region',
                            text="Region parameters require knowledge of image size which is not implemented.")
        pct = self.request.region_pct
        (x,y,w,h)=self.request.region_xywh
        # Convert pct to real
        if (pct):
            x = int( (x / 100.0) * self.width + 0.5)
            y = int( (y / 100.0) * self.height + 0.5)
            w = int( (w / 100.0) * self.width + 0.5)
            h = int( (h / 100.0) * self.height + 0.5)
        # Check if boundary extends beyond image and truncate
        if ((x+w) > self.width):
            w=self.width-x
        if ((y+h) > self.height):
            h=self.height-y
        # Final check to see if we have the whole image
        if ( w==0 or h==0 ):
            raise IIIFError(code=400,parameter='region',
                            text="Region parameters would result in zero size result image.")
        if ( x==0 and y==0 and w==self.width and h==self.height ):
            return(None,None,None,None)
        return(x,y,w,h)

    def size_to_apply(self):
        """Calculate size of image scaled using size parameters

        Assumes current image width and height are available in self.width and 
        self.height, and self.request is IIIFRequest object 

        Formats are: w, ,h w,h pct:p !w,h

        Returns (None,None) if no scaling is required.
        """
        if (self.request.size_full):
            return(None,None)
        elif (self.request.size_pct is not None):
            w = int(self.width * self.request.size_pct / 100.0 + 0.5)
            h = int(self.height * self.request.size_pct / 100.0 + 0.5)
        elif (self.request.size_bang):
            # Have "!w,h" form
            (mw,mh)=self.request.size_wh
            # Pick smaller fraction and then work from that...
            frac = min ( (float(mw)/float(self.width)), (float(mh)/float(self.height)) )
            #print "size=!w,h: mw=%d mh=%d -> frac=%f" % (mw,mh,frac)
            # FIXME - could put in some other function here like factors of two, but
            # FIXME - for now just pick largest image within requested dimensions
            w = int(self.width * frac + 0.5)
            h = int(self.height * frac + 0.5)
        else:
            # Must now be "w,h", "w," or ",h". If both are specified then this will the size,
            # otherwise find other to keep aspect ratio
            (w,h)=self.request.size_wh
            if (w is None):
                w = int(self.width * h / self.height + 0.5)
            elif (h is None):
                h = int(self.height * w / self.width + 0.5)
        # Now have w,h, sanity check and return
        if ( w==0 or h==0 ):
            raise IIIFError(code=400,parameter='size',
                            text="Size parameter would result in zero size result image (%d,%d)."%(w,h))
        # Below would be test for scaling up image size, this is allowed by spec
        # if ( w>self.width or h>self.height ):
        #      raise IIIFError(code=400,parameter='size',
        #                      text="Size requests scaling up image to larger than orginal.")
        if ( w==self.width and h==self.height ):    
            return(None,None)
        return(w,h)

    def rotation_to_apply(self, only90s=False):
        """Check an interpret rotation

        Returns a floating point number 0 <= angle < 360 (degrees).
        """
        rotation=self.request.rotation_deg
        if (only90s and (rotation!=0.0 and rotation!=90.0 and 
                         rotation!=180.0 and rotation!=270.0)):
            raise IIIFError(code=501,parameter="rotation",
                            text="This implementation supports only 0,90,180,270 degree rotations.")
        return(rotation)

    def quality_to_apply(self):
        """Value of color parameter to use in processing request

        Simple substitution of 'native' for default.
        """
        if (self.request.quality is None):
            return('native')
        return(self.request.quality)

    def cleanup(self):
        """Clean up after derive call and use of output

        Call after any output file from the derivation process has been 
        read. Intended to handle any clean up of temporary files or such. 
        This method empty in base class.
        """
        return()



from distutils.core import setup

# Extract version number from resync/_version.py. Here we are very strict
# about the format of the version string as an extra sanity check.
# (thanks for comments in 
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package )
import re
VERSIONFILE="iiif/_version.py"
verfilestr = open(VERSIONFILE, "rt").read()
match = re.search(r"^__version__ = '(\d\.\d.\d+(\.\d+)?)'", verfilestr, re.MULTILINE)
if match:
    version = match.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE))

setup(
    name='iiif',
    version=version,
    author='Simeon Warner',
    author_email='simeon.warner@cornel.edu',
    packages=['iiif'],
    scripts=[],
    classifiers=["Development Status :: 3 - Alpha",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Operating System :: OS Independent", #is this true? know Linux & OS X ok
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 2.6",
                 "Programming Language :: Python :: 2.7",
                 "Topic :: Internet :: WWW/HTTP",
                 "Topic :: Software Development :: Libraries :: Python Modules",
                 "Environment :: Web Environment"],
    url='https://github.com/zimeon/iiif',
    license='LICENSE.txt',
    description='IIIF Image API library',
    long_description=open('README').read(),
)



"""Static file generation for IIIF Image API

Use IIIF Image API manipulations to generate a set of tiles for
a level0 implementation of the IIIF Image API using static files.
"""

import logging
import os
import os.path

from iiif.manipulator_pil import IIIFManipulatorPIL
from iiif.info import IIIFInfo
from iiif.request import IIIFRequest

class IIIFStatic:
    """Provide static generation of IIIF images

    Simplest, using source image as model for directory which
    will be in same directory without extension: 

        IIIFStatic("image1.jpg").generate()

    More complex with different output directory but using name
    derived from image:
    
        sg = IIIFStatic(dst="outdir")
        sg.generate("image2.jpg")
        sg.generate("image3.jpg")
    """

    def __init__(self, src=None, dst=None, prefix=None, tilesize=None, 
                 api_version=None, dryrun=None):
        self.src=src
        self.dst=dst
        self.prefix=prefix
        self.identifier=None
        self.tilesize=tilesize if tilesize is not None else 512
        self.api_version=api_version if api_version is not None else '1.1'
        if (self.api_version=='1'):
            self.apr_version='1.1'
        elif (self.api_version=='2'):
            self.apr_version='2.0'
        self.dryrun = (dryrun is not None)
        self.logger = logging.getLogger('iiif_static')

    def generate(self, src=None, dst=None, tilesize=None):
        # Use params to override object attributes
        if (src is not None):
            self.src=src
        if (dst is not None):
            self.dst=dst
        if (tilesize is not None):
            self.tilesize=tilesize
        # Get image details and calculate tiles
        im=IIIFManipulatorPIL()
        im.srcfile=self.src
        im.do_first()
        width=im.width
        height=im.height
        #print "w=%d h=%d ts=%d" % (im.width,im.height,tilesize)
        xtiles = int(width/self.tilesize)
        ytiles = int(height/self.tilesize)
        max_tiles = xtiles if (xtiles>ytiles) else ytiles
        scale_factors = [ 1 ]
        factor = 1
        for n in range(10):
            if (factor >= max_tiles):
                break
            factor = factor+factor
            scale_factors.append(factor)
        # Setup destination and IIIF identifier
        self.setup_destination(self.src)
        if (self.identifier is None):
            (self.identifier,ext) = os.path.splitext(os.path.basename(self.dst))
        # Write info.json
        info=IIIFInfo(level=0, identifier=self.identifier,
                      width=width, height=height, scale_factors=scale_factors,
                      tile_width=self.tilesize, tile_height=self.tilesize,
                      formats=['jpg'], qualities=['native'],
                      api_version=self.api_version)
        json_file=os.path.join(self.dst,self.identifier,'info.json')
        if (self.dryrun):
            print "dryrun mode, would write the following files:"
            print "%s / %s" % (self.dst, os.path.join(self.identifier,'info.json'))
            self.logger.info(info.as_json())
        else:
            with open(json_file,'w') as f:
                f.write(info.as_json())
                f.close()
            self.logger.info("Written %s"%(json_file))
        # Write out images
        for sf in scale_factors:
            rts = self.tilesize*sf #tile size in original region
            xt = (width-1)/rts+1 
            yt = (height-1)/rts+1
            for nx in range(xt):
                rx = nx*rts
                rxe = rx+rts
                if (rxe>width):
                    rxe=width
                rw = rxe-rx
                sw = rw/sf
                for ny in range(yt):
                    ry = ny*rts
                    rye = ry+rts
                    if (rye>height):
                        rye=height
                    rh = rye-ry
                    sh = rh/sf
                    self.generate_tile([rx,ry,rw,rh],[sw,sh])
        # Now generate reduced size versions of full image
        # 
        # FIXME - Not sure what correct algorithm is for this, from
        # observation of Openseadragon it seems that one keeps halving
        # the pixel size of the full until until both width and height
        # are less than the tile size. The output that tile and further
        # halvings for some time. It seems that without this Openseadragon
        # will not display any unzoomed image in small windows.
        #
        # I do not understand the algorithm that openseadragon uses (or
        # know where it is in the code) to decide how small a version of
        # the complete image to request. It seems that there is a bug in
        # openseadragon here because in some cases it requests images
        # of size 1,1 multiple times. For now, just go all the way down to 
        # this size
        for level in range(1,20):
            factor = 2.0**level
            sw = int(width/factor + 0.5)
            sh = int(height/factor + 0.5)
            if (sw<self.tilesize and sh<self.tilesize):
                if (sw<1 or sh<1):
                    break
                self.generate_tile('full',[sw,sh])

    def generate_tile(self,region,size):
        r = IIIFRequest(identifier=self.identifier)
        if (region == 'full'):
            r.region_full = True
        else:
            r.region_xywh=region # [rx,ry,rw,rh]
        r.size_wh=size # [sw,sh]
        r.format='jpg'
        path = r.url()
        print "%s / %s" % (self.dst,path)
        # Generate...
        if (not self.dryrun):
            m = IIIFManipulatorPIL()
            m.derive(srcfile=self.src, request=r, outfile=os.path.join(self.dst,path))        

    def setup_destination(self, src):
        if (self.dst is None):
            (self.dst, junk) = os.path.splitext(src)
        if (os.path.isdir(self.dst)):
            # Nothin for now, perhaps should delete?
            pass
        elif (os.path.isfile(self.dst)):
            raise Exception("Can't write to directory %s: a file of that name exists"%(self.dst))
        else:
            os.makedirs(self.dst)
        # Now chop off identifier directory
        (self.dst, self.identifier) = os.path.split(self.dst)
        self.logger.info("Output directory %s" % (self.dst))
